use "StraightforwardMachine.sml";
use "SimplerMachine.sml";
(* CSE 341 AB: Section 3, Spring 2019
   Topics in this file:
   -> mutually recursive datatypes
   -> mutually recursive functions
   -> map, filter, join, bind/flat_map/>>=, foldr, foldl
   -> other folds
   -> revisiting hw1
*)

(* Mutually recursive datatypes *)

(* natural numbers and simple recursion example *)
datatype nat = Zro | Succ of nat

fun int_of_nat n =
  case n of
    Zro => 0
  | Succ n => 1 + (int_of_nat n)

(* building numbers so they alternate even and odd *)
(*
 * Notice the special `and` syntax.
 * We can `and` together as many recursive datatypes as we want.
 *)
datatype even = Zero | Even of odd
and      odd  = Odd of even

val my_even_odd = Odd (Even (Odd (Zero)))

(* here's a somewhat less but still contrived example *)
datatype 'a file = File of string * 'a | Dir of string * 'a files
and      'a files = Nil | Cons of 'a file * 'a files

val my_file_system : even file =
  Dir("root",
    Cons(File("foo", Zero),
    Cons(Dir("Desktop",
          Cons(File("bar", Even(Odd(Zero))),
          Nil)),
    Cons(Dir("Downloads", Nil),
  Nil))))

(* How do we recurse over these data structures? *)
(* Mutually recursive functions *)

(* Again notice the `and` syntax. We can `and` together as many as we want, too. *)
fun int_of_even ev =
  case ev of
    Zero => 0
  | Even od => 1 + (int_of_odd od)
and int_of_odd (Odd ev) = 1 + (int_of_even ev)

fun int_of_even_bad ioo ev =
  case ev of
    Zero => 0
  | Even od => 1 + (ioo od)

fun int_of_odd (Odd ev) = 1 + (int_of_even_bad int_of_odd ev)

val int_of_even_bad_fixed = int_of_even_bad int_of_odd

fun file_names f =
  case f of
    File (s, _) => [s]
  | Dir (s, files) => s :: (files_names files)

and files_names fs =
  case fs of
    Nil => []
  | Cons (f, fs) => (file_names f) @ (files_names fs)

val my_file_names = file_names my_file_system
	
(* Abstract Machines *)

(* 
 * Let's use some higher order functions and modules for an only-somewhat contrived example:
 * Computing arithmetic.
 * This is a simpler version of how programming languages are actually computed.
 *
 * Let's revisit the earlier example we had of exp (this time with subtraction, too):
 *)

datatype exp = Constant of int
             | Negate of exp
             | Add of exp * exp
             | Subtract of exp * exp
             | Multiply of exp * exp

(*
 * We wrote an evaluator for it. In fact it's an *interpreter*.
 * It recurses over the source program to produce a value.
 * Since we just call interpret once and it does the recursion for us,
 *   we call it a *big-step* interpreter.
 *)

fun interpret e =
  case e of
      Constant i => i
    | Negate e => ~ (interpret e)
    | Add (e1, e2) => (interpret e1) + (interpret e2)
    | Subtract (e1, e2) => (interpret e1) - (interpret e2)
    | Multiply (e1, e2) => (interpret e1) * (interpret e2)

(*
 * But as we know, this version is not tail recursive. It can create large stack frames.
 * Instead we can do some preprocessing and convert our arithmetic expressions
 * to a language that is list-like so we can use foldl on it.
 * Converting some language A to some other language B is called *compilation*.
 *)

(* Our target language has constants and labels for different operations. *)
datatype oper = Const of int
              | Neg
              | Plus
              | Minus
              | Times

(* A program is just a list of these operations. *)
type program = oper list

(* There are probably more efficient ways to do this conversion, but the guiding notion
   is that compilation is _cheaper_ than evaluation. *)
fun compile e =
  case e of
    Constant i => [Const i]
  | Negate e => (compile e) @ [Neg]
  | Add (e1, e2) => (compile e1) @ (compile e2) @ [Plus]
  | Subtract (e1, e2) => (compile e1) @ (compile e2) @ [Minus]
  | Multiply (e1, e2) => (compile e1) @ (compile e2) @ [Times]

(* 
 * We now define an *abstract machine* that knows how to run our compiled code.
 * An abstract machine has a *state* and a *step* function that knows how to advance the state.
 * Our state is just a stack that keeps track of intermediate values.
 *)

type value = int
type stack = value list
type state = stack

exception BadInput;

fun unary f stack =
  case stack of
    e::tail => (f e)::tail
  | _ => raise BadInput

(* We need to reverse the arguments, because we push the left value onto the
 * stack before the right value.
 *)
fun binary f stack =
  case stack of
    r::l::tail => (f (l, r))::tail
  | _ => raise BadInput

(* The step function takes an operator and returns a function that tells us how
 * to manipulate the state of the machine.
 * _Alternatively_, because it's curried, we can think of it as taking an operator
 * and a state an returning a new state.
 *
 * This is similar to how we can think of map as a curried function.
 * map has type ('a -> 'b) -> 'a list -> 'b list. We can think of it either as
 * a function that takes a function of type 'a -> 'b and lifts it to a function of
 * type 'a list -> 'b list or as a function that takes a function of type 'a -> 'b
 * and an argument of type 'a list and return a result of type 'b list. With tupled
 * arguments we would have to choose which interpretation is correct, but with 
 * currying the only difference is how many arguments we give it!
 *)

(* Notice this is still an interpreter! It tells us how to recurse over the program
 * to get a result. Now though, each call to `step` runs a _single_ step in the execution.
 * This is called a *small-step* interpreter. It's also probably how you think about
 * programs when they run, one step at a time.
 *)
fun step (oper : oper) : state -> state = (* types annotated for clarity *)
  case oper of
    Const i => (fn s => i::s)
  | Neg => unary (op~)
  | Plus => binary (op+)
  | Minus => binary (op-)
  | Times => binary (op*)

fun uncurry f (x, y) = f x y

(* To evaluate a program, we just fold over it and run step. *)
val eval = List.foldl (uncurry step) []

(* A state isn't a value. We write an extract function that takes a value out of
 * a final state of our program.
 *)
fun extract s =
  case s of
    [v] => v
  | _ => raise BadInput

(* To run the whole thing we compile a source program, evaluate it, and extract
 * the result.
 *)
val run = extract o eval o compile

(* scanl: like foldl, but keeps track of the steps that foldl takes *)

fun max(a,b) = if a > b then a else b;
(* 
List.foldl max 0 [1, 2, 0, 5, 3] = 5
scanl max 0 [1, 2, 0, 5, 3]      = [0,1,2,2,5,5]

List.foldl (fn (x, acc) => x::acc) [] [1, 2, 3, 4, 5] = [5,4,3,2,1]
scanl (fn (x, acc) => x::acc) [] [1, 2, 3, 4, 5]      = [[],[1],[2,1],[3,2,1],[4,3,2,1],[5,4,3,2,1]]
 *)












fun scanl f acc xs =
  case xs of
    [] => [acc]
  | x::xs' => acc::(scanl f (f (x, acc)) xs')

(* Now we have a way to trace the execution of our program! *)
val eval_trace = scanl (uncurry step) []

val run_trace = eval_trace o compile

(* Arithmetic Machine Interface *)

(* imports are at the top of the file *)

(* like type but for modules *)
structure SM = SimplerMachine

(* With some more advanced module features (namely functors), we could even
 * get rid of this boilerplate!
 * In any case, as long as the machine implements the ARITHMETIC_MACHINE interface,
 * we can write these functions.
 *)
val sm_eval = List.foldl (uncurry SM.step) []
val sm_run = SM.extract o sm_eval o SM.compile
val sm_eval_trace = scanl (uncurry SM.step) []
val sm_run_trace = sm_eval_trace o SM.compile

val sm_sub_ex =
  let
    open SM
  in
    Subtract (Constant 5, Constant 6)
  end
